﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Random;
using MathNet.Numerics.Statistics;
using DoubleVector = MathNet.Numerics.LinearAlgebra.Vector<double>;

namespace HeuristicLab.Problems.Instances.DataAnalysis {
  public abstract class VectorDataTestOne : ArtificialRegressionDataDescriptor {

    protected const int Rows = 1000;

    public override string Description { get { return ""; } }

    protected override string TargetVariable { get { return "Y"; } }
    protected override string[] VariableNames { get { return new string[] { "X1", "X2", "X3", "V1", "V2", "Y" }; } }
    protected override string[] AllowedInputVariables { get { return new string[] { "X1", "X2", "X3", "V1", "V2" }; } }
    protected override int TrainingPartitionStart { get { return 0; } }
    protected override int TrainingPartitionEnd { get { return Rows * 3 / 4; } }
    protected override int TestPartitionStart { get { return TrainingPartitionEnd; } }
    protected override int TestPartitionEnd { get { return Rows; } }

    public int Seed { get; private set; }

    protected VectorDataTestOne()
      : this((int)DateTime.Now.Ticks) { }
    protected VectorDataTestOne(int seed)
      : base() {
      Seed = seed;
    }


    protected override List<List<double>> GenerateValues() { return null; }
    protected override List<IList> GenerateValuesExtended() {
      var rand = new MersenneTwister((uint)Seed);

      double x1, x2, x3;
      DoubleVector v1, v2;
      double y;

      var x1Column = new List<double>(Rows);
      var x2Column = new List<double>(Rows);
      var x3Column = new List<double>(Rows);
      var v1Column = new List<DoubleVector>(Rows);
      var v2Column = new List<DoubleVector>(Rows);
      var yColumn = new List<double>(Rows);

      var vectorLengths = GetVectorLengths(rand);
      for (int i = 0; i < Rows; i++) {
        x1 = rand.NextDouble(-2, 2);
        x2 = rand.NextDouble(2, 6);
        x3 = rand.NextDouble(0, 1);
        int v1Length = vectorLengths[0][i];
        int v2Length = vectorLengths[1][i];
        v1 = rand.NextDoubleVector(2, 6, v1Length);
        v2 = rand.NextDoubleVector(3, 5, v2Length);

        y = x1 * v1.Sum() + x2 * v2.Mean();

        x1Column.Add(x1);
        x2Column.Add(x2);
        x3Column.Add(x3);
        v1Column.Add(v1);
        v2Column.Add(v2);
        yColumn.Add(y);
      }

      return new List<IList> { x1Column, x2Column, x3Column, v1Column, v2Column, yColumn };
    }

    protected abstract List<int>[] GetVectorLengths(IRandom rand);
  }

  public class VectorDataTestOneA : VectorDataTestOne {
    public override string Name {
      get { return "Vector Data Test - I [fully-constrained]: Y = X1 * sum(V1) + X2 * mean(V2)"; }
    }

    public VectorDataTestOneA() : base() { }
    public VectorDataTestOneA(int seed) : base(seed) { }

    protected override List<int>[] GetVectorLengths(IRandom rand) {
      // always same length
      const int length = 5;
      return new List<int>[2] {
        Enumerable.Repeat(length, Rows).ToList(),
        Enumerable.Repeat(length, Rows).ToList()
      };
    }
  }

  public class VectorDataTestOneB : VectorDataTestOne {
    public override string Name { get { return "Vector Data Test - I [row-constrained]: Y = X1 * sum(V1) + X2 * mean(V2)"; } }

    public VectorDataTestOneB() : base() { }
    public VectorDataTestOneB(int seed) : base(seed) { }

    protected override List<int>[] GetVectorLengths(IRandom rand) {
      // length between length 4 and 8, same row always the same length
      var lengths = Enumerable.Range(0, Rows).Select(i => rand.Next(4, 8)).ToList();
      return new List<int>[2] {
        Enumerable.Range(0, Rows).Select(i => lengths[i]).ToList(),
        Enumerable.Range(0, Rows).Select(i => lengths[i]).ToList()
      };
    }
  }

  public class VectorDataTestOneC : VectorDataTestOne {
    public override string Name { get { return "Vector Data Test - I [column-constrained]: Y = X1 * sum(V1) + X2 * mean(V2)"; } }

    public VectorDataTestOneC() : base() { }
    public VectorDataTestOneC(int seed) : base(seed) { }

    protected override List<int>[] GetVectorLengths(IRandom rand) {
      // length between length 4 and 8; each feature is same length
      // force two different lengths
      int v1Length = rand.Next(4, 8);
      int v2Length;
      do {
        v2Length = rand.Next(4, 8);
      } while (v1Length != v2Length);
      return new List<int>[2] {
        Enumerable.Repeat(v1Length, Rows).ToList(),
        Enumerable.Repeat(v2Length, Rows).ToList()
      };
    }
  }

  public class VectorDataTestOneD : VectorDataTestOne {
    public override string Name { get { return "Vector Data Test - I [unconstrained]: Y = X1 * sum(V1) + X2 * mean(V2)"; } }

    public VectorDataTestOneD() : base() { }
    public VectorDataTestOneD(int seed) : base(seed) { }

    protected override List<int>[] GetVectorLengths(IRandom rand) {
      // always random between 4 and 8
      return new List<int>[2] {
          Enumerable.Range(0, Rows).Select(i => rand.Next(4, 8)).ToList(),
          Enumerable.Range(0, Rows).Select(i => rand.Next(4, 8)).ToList()
        };
    }
  }
}